
static final int WORLD_EXTENT = 256;
static final int WORLD_HEIGHT = 128;

static final float DRAW_DISTANCE = 48.f;
static final int CLOUD_DISTANCE = ceil(DRAW_DISTANCE);

static final float MOVE_SPEED = 0.05, TURN_SPEED = 0.05; 

static final int STONE = 1, DIRT = 2, GRASS = 3, BEDROCK = 17, LOG = 20, LEAVES = 53, DIAMOND_ORE = 50, IRON_ORE = 33, COAL_ORE = 34;

import java.util.HashSet;
import java.util.Random;

HashSet<Integer> held_keys = new HashSet<>();

void keyPressed() {
  held_keys.add(keyCode);
  if (keyCode == 99) debugInfo = !debugInfo;
  if (keyCode == 100) noclip = !noclip;
}
void keyReleased() {
  held_keys.remove(keyCode);
}
void mouseDragged() {
  playerPitch += (mouseX - pmouseX) * 0.006f;
  playerYaw += (mouseY - pmouseY) * 0.006f;
}

byte[] blocks;
byte[] height_map;
boolean[] cloud_array;

float playerX,playerY = 72.f,playerZ = -2.f;
float velX,velY,velZ;
float playerPitch,playerYaw;
boolean noclip = false, debugInfo = true;

float cloud_position;

PImage terrain;

int getBlock(int x,int y,int z) {
  x += WORLD_EXTENT/2;
  z += WORLD_EXTENT/2;
  if (x < 0 || x >= WORLD_EXTENT || z < 0 || z >= WORLD_EXTENT || y < 0 || y >= WORLD_HEIGHT) return 0;
  return blocks[y+x*WORLD_HEIGHT+z*(WORLD_EXTENT*WORLD_HEIGHT)]&0xFF;
}
void setBlock(int x,int y,int z, int blkid) {
  x += WORLD_EXTENT/2;
  z += WORLD_EXTENT/2;
  if (x < 0 || x >= WORLD_EXTENT || z < 0 || z >= WORLD_EXTENT || y < 0 || y >= WORLD_HEIGHT) return;
  blocks[y+x*WORLD_HEIGHT+z*(WORLD_EXTENT*WORLD_HEIGHT)] = (byte)blkid;
  int h = height_map[x+z*WORLD_EXTENT];
  if (blkid != 0 && y > h) {
    height_map[x+z*WORLD_EXTENT] = (byte)y;
  } else if (blkid == 0 && y == h) {
    for (;h>=0;h--) if (getBlock(x,h,z)!=0) break;
    if (h<0) h = 0;
    height_map[x+z*WORLD_EXTENT] = (byte)h;
  }
}
void setBlockSoft(int x,int y,int z, int blkid) {
  if(getBlock(x,y,z)==0)setBlock(x,y,z,blkid);
}


int getHeight(int x, int z) {
  x += WORLD_EXTENT/2;
  z += WORLD_EXTENT/2;
  if (x < 0 || x >= WORLD_EXTENT || z < 0 || z >= WORLD_EXTENT) return 0;
  return height_map[x+z*WORLD_EXTENT]&0xFF;
}


float fsqr(float x) {
  return x*x;
}


int hash(int x) {
    x = ((x >>> 16) ^ x) * 0x45d9f3b;
    x = ((x >>> 16) ^ x) * 0x45d9f3b;
    x = (x >>> 16) ^ x;
    return x;
}

void setup() {
  hint(DISABLE_TEXTURE_MIPMAPS);
  noSmooth();
  size(1024,768,P3D); 
  terrain = loadImage("terrain.png");
  for(int x=0;x<16;x++)for(int y=0;y<16;y++) terrain.set(x,y,PImage.blendColor(terrain.get(x,y),#99FF77,MULTIPLY));
  for(int x=80;x<96;x++)for(int y=48;y<64;y++) terrain.set(x,y,PImage.blendColor(terrain.get(x,y),#99FF77,MULTIPLY));
  ((PGraphicsOpenGL)g).textureSampling(2); // HACK
  levelGenerate();
}


void treeAt(int x, int y, int z) {
  setBlock(x,y+0,z,LOG);
  setBlock(x,y+1,z,LOG);
  
  setBlock(x,y+2,z,LOG);
  for(int x2=x-1;x2<=x+1;x2++)for(int z2=z-1;z2<=z+1;z2++) setBlockSoft(x2,y+2,z2,LEAVES);
  
  setBlock(x,y+3,z,LOG);
  for(int x2=x-2;x2<=x+2;x2++)for(int z2=z-2;z2<=z+2;z2++) setBlockSoft(x2,y+3,z2,LEAVES);
  
  setBlock(x,y+4,z,LOG);
  for(int x2=x-2;x2<=x+2;x2++)for(int z2=z-2;z2<=z+2;z2++) setBlockSoft(x2,y+4,z2,LEAVES);
  
  for(int x2=x-1;x2<=x+1;x2++)for(int z2=z-1;z2<=z+1;z2++) setBlockSoft(x2,y+5,z2,LEAVES);
}

void levelGenerate() {
  blocks = new byte[WORLD_EXTENT*WORLD_EXTENT*WORLD_HEIGHT];
  height_map = new byte[WORLD_EXTENT*WORLD_EXTENT];
  cloud_array = new boolean[256*256];
  
  for (int x=0;x<256;x++) {
    for (int y=0;y<256;y++) {
      cloud_array[x+y*256] = ImprovedNoise.noise(x/6.f,y/6.f,0) > 0.1;
    }
  }
  
  
  // build 3D terrain
  for (int x=-WORLD_EXTENT/2;x<WORLD_EXTENT/2;x++) {
    for (int z=-WORLD_EXTENT/2;z<WORLD_EXTENT/2;z++) {
      int blocks_set = 0;
      for (int y=WORLD_HEIGHT-1;y>=3;y--) {
        float threshold = (y - 64.f) / 64.f;
        float density = 0;
        density += (float)ImprovedNoise.noise(x/64.f,y/64.f,z/64.f) / 2;
        density += (float)ImprovedNoise.noise(x/32.f,y/32.f,z/32.f) / 4;
        density += (float)ImprovedNoise.noise(x/16.f,y/16.f,z/16.f) / 8;
        density += (float)ImprovedNoise.noise(x/8.f,y/8.f,z/8.f) / 16;
        
        if (density > threshold) {
          if (blocks_set == 0) setBlock(x,y,z,GRASS);
          else if (blocks_set < 4) setBlock(x,y,z,DIRT);
          else setBlock(x,y,z,STONE);
          blocks_set++;
        }
      setBlock(x,2,z,(hash(x+z*512)&3) == 0 ? BEDROCK : STONE);
      setBlock(x,1,z,(hash(x+z*512)&1) == 0 ? BEDROCK : STONE);
      setBlock(x,0,z,BEDROCK);
        
      }
    }
  }
  
  // plant trees
  int numTrees = WORLD_EXTENT*WORLD_EXTENT / 100;
  Random r = new Random(123);
  for(int i=0;i<numTrees;i++) {
    int x = r.nextInt(WORLD_EXTENT) - WORLD_EXTENT/2;
    int z = r.nextInt(WORLD_EXTENT) - WORLD_EXTENT/2;
    int ground = getHeight(x,z);
    if (ground < WORLD_HEIGHT - 8 && getBlock(x,ground,z) == GRASS) {
      treeAt(x,ground+1,z);
      setBlock(x,ground,z,DIRT);
    }
  }
  
  // place ores
  r = new Random(1234);
  int numDiamonds = WORLD_EXTENT*WORLD_EXTENT / 20;
  for (int i=0;i<numDiamonds;i++) {
    int x = r.nextInt(WORLD_EXTENT) - WORLD_EXTENT/2;
    int z = r.nextInt(WORLD_EXTENT) - WORLD_EXTENT/2;
    int y = 5 + (r.nextInt(10) + r.nextInt(10))/2;
    if (getBlock(x,y,z) == STONE) setBlock(x,y,z,DIAMOND_ORE);
  }
  int numIron = WORLD_EXTENT*WORLD_EXTENT / 4;
  for (int i=0;i<numIron;i++) {
    int x = r.nextInt(WORLD_EXTENT) - WORLD_EXTENT/2;
    int z = r.nextInt(WORLD_EXTENT) - WORLD_EXTENT/2;
    int y = 12 + (r.nextInt(40) + r.nextInt(40))/2;
    if (getBlock(x,y,z) == STONE) setBlock(x,y,z,IRON_ORE);
  }
  int numCoal = WORLD_EXTENT*WORLD_EXTENT * 1;
  for (int i=0;i<numCoal;i++) {
    int x = r.nextInt(WORLD_EXTENT) - WORLD_EXTENT/2;
    int z = r.nextInt(WORLD_EXTENT) - WORLD_EXTENT/2;
    int y = 32 + (r.nextInt(80) + r.nextInt(80))/2;
    if (getBlock(x,y,z) == STONE) setBlock(x,y,z,COAL_ORE);
  }
  
  
  
  setBlock(0,0,0,0);
  setBlock(0,1,0,0);
  setBlock(0,2,0,0);
  setBlock(0,3,0,0);
  setBlock(0,4,0,0);
  setBlock(0,5,0,0);
  setBlock(0,6,0,0);
  setBlock(0,7,0,0);
  setBlock(0,8,0,0);
}


int topFace(int blkid) {
  switch (blkid) {
    case GRASS: return 0;
    case LOG: return LOG+1;
    default: return blkid; 
  }
}
int bottomFace(int blkid) {
  switch (blkid) {
    case GRASS: return DIRT;
    case LOG: return LOG+1;
    default: return blkid; 
  }
}

void draw() {
  
  float movesp = MOVE_SPEED;
  if (held_keys.contains((int)'E')) movesp *= 4;
  
  if (held_keys.contains((int)'W')) {
    playerX += sin(playerPitch)*movesp;
    playerZ += cos(playerPitch)*movesp;
  }
  if (held_keys.contains((int)'S')) {
    playerX -= sin(playerPitch)*movesp;
    playerZ -= cos(playerPitch)*movesp;
  }
  if (held_keys.contains((int)'A')) {
    playerX -= cos(playerPitch)*movesp;
    playerZ += sin(playerPitch)*movesp;
  }
  if (held_keys.contains((int)'D')) {
    playerX += cos(playerPitch)*movesp;
    playerZ -= sin(playerPitch)*movesp;
  }
  if (noclip && held_keys.contains((int)' ')) playerY += movesp;
  if (held_keys.contains(SHIFT)) playerY -= movesp;
  if (held_keys.contains(LEFT)) playerPitch -= TURN_SPEED;
  if (held_keys.contains(RIGHT)) playerPitch += TURN_SPEED;
  if (held_keys.contains(UP)) playerYaw -= TURN_SPEED;
  if (held_keys.contains(DOWN)) playerYaw += TURN_SPEED;
  
  if (!noclip) velY -= 0.01; // Gravity?
  
  playerX += velX;
  playerY += velY;
  playerZ += velZ;
  
  
  if (!noclip) {
    int dumb = getHeight((int)floor(playerX),(int)floor(playerZ))+1;
    if (dumb >= playerY) {
      velY = 0;
      playerY = dumb;
      if (held_keys.contains((int)' ')) velY = (held_keys.contains((int)'E')) ? 0.7f : 0.2f;
      velX *= 0.6;
      velZ *= 0.6;
    } else {
      velX *= 0.99;
      velY *= 0.99;
      velZ *= 0.99;
    }
  } else { 
      velX *= 0.6;
      velY *= 0.6;
      velZ *= 0.6;
  }
  
  float eyeX = playerX, eyeY = playerY + 1.4f, eyeZ = playerZ;
  float eyePitch = playerPitch, eyeYaw = playerYaw;
  
  background(128,128,200);
  beginCamera();
  perspective(80,float(width)/float(height),0.1,1000.0);
  camera(0,0,0, 0,0,1,  0.f,1.f,0.f);
  rotateX(-eyeYaw);
  rotateY(-eyePitch);
  translate(-eyeX,-eyeY,-eyeZ);  
  
  
  noStroke();

  
  
  
  beginShape(QUADS);
  textureMode(NORMAL);
  texture(terrain);
  for (int x=-WORLD_EXTENT/2;x<WORLD_EXTENT/2;x++) {
    for (int z=-WORLD_EXTENT/2;z<WORLD_EXTENT/2;z++) {
      if (fsqr(x-eyeX)+fsqr(z-eyeZ) > fsqr(DRAW_DISTANCE))continue;
      for (int y=0;y<WORLD_HEIGHT;y++) {
        int block = getBlock(x,y,z);
        if (block==0) continue;
        
        if (getBlock(x,y-1,z)==0) {
          tint(92); // Must be dark
          int texid = bottomFace(block);
          float u = texU(texid), v = texV(texid);
          vertex(x  ,y  ,z+1,u       ,v       );
          vertex(x+1,y  ,z+1,u+UVSTEP,v       );
          vertex(x+1,y  ,z  ,u+UVSTEP,v+UVSTEP);
          vertex(x  ,y  ,z  ,u       ,v+UVSTEP);
        }
        if (getBlock(x,y+1,z)==0) {
          tint(getHeight(x,z) <= y ? 255:128);
          int texid = topFace(block);
          float u = texU(texid), v = texV(texid);
          vertex(x  ,y+1,z  ,u       ,v       );
          vertex(x+1,y+1,z  ,u+UVSTEP,v       );
          vertex(x+1,y+1,z+1,u+UVSTEP,v+UVSTEP);
          vertex(x  ,y+1,z+1,u       ,v+UVSTEP);
        }
        if (getBlock(x-1,y,z)==0) {
          tint(getHeight(x-1,z) <= y ? 200:100);
          int texid = block;
          float u = texU(texid), v = texV(texid);
          vertex(x  ,y+1,z  ,u+UVSTEP,v       );
          vertex(x  ,y+1,z+1,u       ,v       );
          vertex(x  ,y  ,z+1,u       ,v+UVSTEP);
          vertex(x  ,y  ,z  ,u+UVSTEP,v+UVSTEP);
        }
        if (getBlock(x+1,y,z)==0) {
          tint(getHeight(x+1,z) <= y ? 200:100);
          int texid = block;
          float u = texU(texid), v = texV(texid);
          vertex(x+1,y  ,z  ,u       ,v+UVSTEP);
          vertex(x+1,y  ,z+1,u+UVSTEP,v+UVSTEP);
          vertex(x+1,y+1,z+1,u+UVSTEP,v       );
          vertex(x+1,y+1,z  ,u       ,v       );
        }
        if (getBlock(x,y,z-1)==0) {
          tint(getHeight(x,z-1) <= y ? 248:124);
          int texid = block;
          float u = texU(texid), v = texV(texid);
          vertex(x  ,y+1,z  ,u       ,v       );
          vertex(x+1,y+1,z  ,u+UVSTEP,v       );
          vertex(x+1,y  ,z  ,u+UVSTEP,v+UVSTEP);
          vertex(x  ,y  ,z  ,u       ,v+UVSTEP);
        }
        if (getBlock(x,y,z+1)==0) {
          tint(getHeight(x,z+1) <= y ? 248:124);
          int texid = block;
          float u = texU(texid), v = texV(texid);
          vertex(x  ,y  ,z+1,u+UVSTEP,v+UVSTEP);
          vertex(x+1,y  ,z+1,u       ,v+UVSTEP);
          vertex(x+1,y+1,z+1,u       ,v       );
          vertex(x  ,y+1,z+1,u+UVSTEP,v       );
        }
        
        
        
      }
    }
  }
  endShape();
  
  // Draw sun
  fill(255,255,128,255);
  beginShape(QUADS);
  vertex(eyeX-100,eyeY+800,eyeZ-100);
  vertex(eyeX+100,eyeY+800,eyeZ-100);
  vertex(eyeX+100,eyeY+800,eyeZ+100);
  vertex(eyeX-100,eyeY+800,eyeZ+100);
  endShape();
  
    
  // Draw clouds
  float cloud_height = 150.f;
  int cx_mid = ((int)floor(eyeX))/8;
  int cz_mid = ((int)floor(eyeZ))/8;
  fill(255,255,255,128);
  beginShape(QUADS);
  for(int x=-CLOUD_DISTANCE;x<=CLOUD_DISTANCE;x++) {
    for(int z=-CLOUD_DISTANCE;z<=CLOUD_DISTANCE;z++) {
      int tilex = (cx_mid+x)&255, tilez = (cz_mid+z+(int)floor(cloud_position))&255;
      if (cloud_array[tilex+tilez*256]) {
        float offset = 8*(floor(cloud_position) - cloud_position); 
        float x1 = (cx_mid+x)*8, x2 = (cx_mid+x+1)*8;
        float z1 = cz_mid*8+max((z)*8+offset,-CLOUD_DISTANCE*8), z2 = cz_mid*8+min((z+1)*8+offset,CLOUD_DISTANCE*8);
        vertex(x1,cloud_height,z1);
        vertex(x2,cloud_height,z1);
        vertex(x2,cloud_height,z2);
        vertex(x1,cloud_height,z2);
      }
      
    }
  }  
  endShape();
  cloud_position += 0.03f;
  
  
  resetMatrix();
  endCamera();
  ortho();
  translate(-width/2, -height/2, 0);

  fill(255,255,255);
  textSize(16);
  text("Fakecraft Indev",0,16*1);
  if (noclip) text("NoClip!",200,16*1); 
  if (debugInfo) {
    text(String.format("X: %.6f",playerX),0,16*2);
    text(String.format("Y: %.6f",playerY),0,16*3);
    text(String.format("Z: %.6f",playerZ),0,16*4);
  }
  
}

static final float UVSTEP = 1.f/16 - 0.002;

float texU(int texId) {
  return (texId&0xF)*(1.f/16) + 0.001;
}
float texV(int texId) {
  return (texId>>4)*(1.f/16) + 0.001;
}
